using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using UnityEngine.Events;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;

namespace Stopper {
	public class AdditionalControllers : MVRScript {
		// Prefabricated UI for joint controllers. Not sure how to get ahold of this
		// other than copying from somewhere
		GameObject controllerUIPrefab;
		
		// A sample controller for retrieving controller meshes. Again, not sure
		// how to find this other than copying from elsewhere
		FreeControllerV3 sampleController;
		
		Atom mainAtom;
		
		GameObject jointControl;
		GameObject ui;
		
		GameObject uiPlaceholder;
		
		bool addingToAtom = false;
		List<FreeControllerV3> createdControllers = new List<FreeControllerV3>();
		
		// Rigidbodies to add controllers for. 
		List<string> rigidBodyControllerList = new List<string> { 
			"tongueTip",
			"rIndex3", "lIndex3",
			"rMid3", "lMid3",
			"rRing3", "lRing3",
			"rPinky3", "lPinky3",
			"rThumb3", "lThumb3",
		};  
		
		// Plugin UI
		JSONStorableFloat newControllerSelectRadius;
		JSONStorableFloat handSelectRadius;
		JSONStorableFloat allSelectRadius;
		
		public override void Init() {
			CreateUI();
		}
		
		private void CreateUI(){
			newControllerSelectRadius = AddSlider("Added Controller Select Radius", 0.01f, 0.01f, 0.1f, delegate {SyncControllerSelectRadius();});
			handSelectRadius = AddSlider("Hand Select Radius", 0.03f, 0.01f, 0.1f, delegate {SyncControllerSelectRadius();});
			allSelectRadius = AddSlider("All Controller Select Radius", 0.1f, 0.01f, 0.1f, delegate {SyncControllerSelectRadius();});
		}
		
		private void SyncControllerSelectRadius(){
			if(jointControl != null){
				foreach(Transform childTransform in jointControl.transform){
					GameObject control = childTransform.gameObject;
					
					float selectRadius = allSelectRadius.val;
					
					if(control.name.Contains("Hand")){
						selectRadius = handSelectRadius.val;
					}
					
					foreach(string addedControlName in rigidBodyControllerList){
						if(control.name.Contains(addedControlName)){
							selectRadius = newControllerSelectRadius.val;
							break;
						}
					}
					
					control.GetComponent<SphereCollider>().radius = selectRadius;
				}
			}
		}
		
		private GameObject FindGameObject(GameObject parent, string name){
			Transform childTransform = parent.transform.Find(name);
			
			if(childTransform == null){
				return null;
			} else {
				return childTransform.gameObject;
			}
		}
		
		private GameObject GetOrCreateChildObject(GameObject parent, string name) {
			GameObject gameObject = FindGameObject(parent, name);
			if(gameObject == null) {
				gameObject = new GameObject(name);
				gameObject.transform.parent = parent.transform;
			} 
			
			return gameObject;
		}
		
		private Component GetOrCreateComponent(GameObject parent, System.Type type) {
			Component component = parent.GetComponent(type) as Component;
			if(component == null){
				component = parent.AddComponent(type) as Component;
			}
			return component;
		}
		
		List<GameObject> GetAllChildren(GameObject parent, List<GameObject> objectList = null)
		{
			if (objectList == null) objectList = new List<GameObject>();
		  
			foreach (Transform child in parent.transform) {
				objectList.Add(child.gameObject);
				GetAllChildren(child.gameObject, objectList);
			}
			return objectList;
		}
		
		//-------Start------
		private void Start()
		{
			StartCoroutine("AddControllers");
		}
		
		private IEnumerator AddControllers()
		{
			/* 
				When added as a session plugin, this script works by modifying the Person prefab - the template from which other People are created.
				
				Why? Because Atoms have an immutable freeControllers array that's used for various things like 'select
				collider from scene'. 
				
				If that doesn't matter to you, set mainAtom = containingAtom and delete all the prefab code.			
			*/
			if(containingAtom.name == "CoreControl"){
				
				// Check to see if the prefab already exists (there's a person in the scene)
				GameObject personPrefab = SuperController.singleton.GetCachedPrefab("a_per", "Person");
				if(personPrefab == null){
					// No people in the scene yet. Load one (without adding to scene) and register it as the prefab. 
					// This might have unintended consequences. 
					SuperController.LogMessage("Person prefab not cached. Loading from asset bundle.");
					AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync("a_per", "Person", typeof(GameObject));
					if (request == null)
					{
						SuperController.LogError("Unable to load Person prefab");
						yield break;
					}
					yield return base.StartCoroutine(request);

					personPrefab = request.GetAsset<GameObject>();
					if (personPrefab != null)
					{
						SuperController.singleton.RegisterPrefab("a_per", "Person", personPrefab);
						mainAtom = personPrefab.GetComponent<Atom>();
					} else {
						SuperController.LogError("Unable to load Person prefab");
						yield break;
					}
				} else {
					mainAtom = personPrefab.GetComponent<Atom>();
				}
			} else {
				// If we didn't add it as a session atom, just apply it to the current atom.
				addingToAtom = true;
				mainAtom = containingAtom;
			}
			
			GameObject sampleHipControl = null;
			
			// Load an existing controller joint ui to use as a template. Not sure how else to do it.
			foreach (GameObject go in Resources.FindObjectsOfTypeAll(typeof(GameObject)) as GameObject[]){
				if(go.name =="UIFreeControllerJoint") {
					controllerUIPrefab = go;
					continue;
				}
			}
			
			foreach (GameObject go in GetAllChildren(mainAtom.gameObject)){
				if(go.name == "hipControl") {
					sampleHipControl = go;
				} else if (go.name == "JointControl") {
					jointControl = go;
				} else if (go.name == "UI" && go.transform.parent.gameObject == mainAtom.gameObject) {
					ui = go;
				}
			}
			
			sampleController = sampleHipControl.GetComponent<FreeControllerV3>();
			
			foreach(Rigidbody rb in mainAtom.gameObject.GetComponentsInChildren<Rigidbody>()){
				//if(rb.name.Contains("tongue")){
				//rb.solverVelocityIterations = 10;
				//}
				
				if(!rigidBodyControllerList.Contains(rb.name)){
					continue;
				}
				
				bool previouslyLoaded = FindGameObject(jointControl, rb.name+"Control") != null;
				
				if(previouslyLoaded){
					continue;
				}
				
				rb.gameObject.SetActive(false);
				
				// Parent object for controls
				//  Needs a Rigidbody, SphereCollider, FreeControllerV3, and a joint tying it to the target Rigidbody
				GameObject control = GetOrCreateChildObject(jointControl, rb.name+"Control");
				control.SetActive(false);
				control.layer = 8;
				
				ForceReceiver forceReceiver = GetOrCreateComponent(rb.gameObject, typeof(ForceReceiver)) as ForceReceiver;
				forceReceiver.containingAtom = mainAtom;

				Rigidbody controlRB = GetOrCreateComponent(control, typeof(Rigidbody)) as Rigidbody;
				controlRB.isKinematic = false;
				controlRB.useGravity = false;
				controlRB.detectCollisions = true;
				
				ConfigurableJoint controlJoint = null;

				foreach(ConfigurableJoint joint in rb.gameObject.GetComponents<ConfigurableJoint>()){
					if(joint.connectedBody == controlRB){
						controlJoint = joint;
					}
				}
				
				if(controlJoint == null){
					controlJoint = rb.gameObject.AddComponent(typeof(ConfigurableJoint)) as ConfigurableJoint;
				}
				
				
				controlJoint.connectedBody = controlRB;
				controlJoint.autoConfigureConnectedAnchor = false;
				
				FreeControllerV3 controller = GetOrCreateComponent(control, typeof(FreeControllerV3)) as FreeControllerV3;
				controller.containingAtom = mainAtom;
				controller.followWhenOffRB = rb;
				controller.useGravityOnRBWhenOff = false;
				controller.meshScale = 0.2f;
				controller.enableSelectRoot = true;
				controller.enabled = true;
				controller.hidden = false;
				controller.startingPositionState = FreeControllerV3.PositionState.Off;
				controller.startingRotationState = FreeControllerV3.RotationState.Off;
				
				createdControllers.Add(controller);
				
				// Add it to the SuperController's free controller list so that global enable/disable works
				if(addingToAtom && !SuperController.singleton.GetAllFreeControllers().Contains(controller)){
					SuperController.singleton.GetAllFreeControllers().Add(controller);
				}
				
				controller.material = sampleController.material;
				if(controller.onPositionMesh == null){
					controller.onPositionMesh = sampleController.onPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.onPositionMesh, controller.transform);
					controller.offPositionMesh = sampleController.offPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.offPositionMesh, controller.transform);
					controller.followingPositionMesh = sampleController.followingPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.followingPositionMesh, controller.transform);
					controller.holdPositionMesh = sampleController.holdPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.holdPositionMesh, controller.transform);
					controller.lockPositionMesh = sampleController.lockPositionMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lockPositionMesh, controller.transform);
					controller.onRotationMesh = sampleController.onRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.onRotationMesh, controller.transform);
					controller.offRotationMesh = sampleController.offRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.offRotationMesh, controller.transform);
					controller.followingRotationMesh = sampleController.followingRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.followingRotationMesh, controller.transform);
					controller.holdRotationMesh = sampleController.holdRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.holdRotationMesh, controller.transform);
					
					controller.lockRotationMesh = sampleController.lockRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lockRotationMesh, controller.transform);
					controller.lookAtRotationMesh = sampleController.lookAtRotationMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.lookAtRotationMesh, controller.transform);
					controller.moveModeOverlayMesh = sampleController.moveModeOverlayMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.moveModeOverlayMesh, controller.transform);
					controller.rotateModeOverlayMesh = sampleController.rotateModeOverlayMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.rotateModeOverlayMesh, controller.transform);
					controller.deselectedMesh = sampleController.deselectedMesh == null ? null :  UnityEngine.Object.Instantiate<Mesh>(sampleController.deselectedMesh, controller.transform);;
				}
				
				SphereCollider collider = GetOrCreateComponent(control, typeof(SphereCollider)) as SphereCollider;
				collider.enabled = true;
				collider.radius = 0.01f;
				
				// Controller Link. Not sure what this is used for.
				GameObject controlLink = GetOrCreateChildObject(jointControl, rb.name+"Link");
				controlLink.layer = 8;
				
				SphereCollider linkCollider = GetOrCreateComponent(controlLink, typeof(SphereCollider)) as SphereCollider;
				linkCollider.enabled = true;
				linkCollider.radius = 0.01f;
				
				Rigidbody linkRB = GetOrCreateComponent(controlLink, typeof(Rigidbody)) as Rigidbody;
				linkRB.isKinematic = true;
				linkRB.useGravity = false;
				linkRB.detectCollisions = false;
				
				FreeControllerV3Link freeControllerLink = GetOrCreateComponent(controlLink, typeof(FreeControllerV3Link)) as FreeControllerV3Link;
				freeControllerLink.linkedController = controller;
				
				// UI			
				uiPlaceholder = GetOrCreateChildObject(ui, "UIPlaceholder"+rb.name );
				uiPlaceholder.SetActive(false);
				
				MoveAndRotateAsHUDAnchor moveAsHud = GetOrCreateComponent(uiPlaceholder, typeof(MoveAndRotateAsHUDAnchor)) as MoveAndRotateAsHUDAnchor;
				moveAsHud.scaleAsAnchor = true;
				moveAsHud.anchorNum = HUDAnchor.AnchorNum.Two;
									
				UIConnectorMaster connectorMaster = GetOrCreateComponent(uiPlaceholder, typeof(UIConnectorMaster)) as UIConnectorMaster;
				connectorMaster.UIPrefab = controllerUIPrefab.transform;
						
				UIConnector connector = GetOrCreateComponent(uiPlaceholder, typeof(UIConnector)) as UIConnector;
				connector.receiver = controller;
				connector.receiverTransform = control.transform;
				
				controller.UITransforms = new Transform[]{uiPlaceholder.transform};
				controller.UITransformsPlayMode = new Transform[]{};
				uiPlaceholder.SetActive(true);
				rb.gameObject.SetActive(true);
				control.SetActive(true);
				
				
				// Needs to be done here or else things break. not sure why
				controller.guihidden = true;
			}
			
			// Add the new controllers to the AllJointsController so they work with the Key/All joint controls
			AllJointsController allJointsController = jointControl.GetComponent<AllJointsController>();
			List<FreeControllerV3> freeControllers = new List<FreeControllerV3>();
			freeControllers.AddRange(allJointsController.freeControllers);
			
			foreach(FreeControllerV3 newController in createdControllers){
				if(!freeControllers.Contains(newController)){
					freeControllers.Add(newController);
				}
			}
			allJointsController.freeControllers = freeControllers.ToArray();
			
			SyncControllerSelectRadius();
		}
		
		private void OnDisable()
		{
			// Clean up maybe? Seems like a hassle.
		}
		
		//UI Functions
		
		public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, UnityAction<float> onChange = null, bool rightSide = false)
		{
			JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
			storableFloat.storeType = JSONStorableParam.StoreType.Full;
			UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
			RegisterFloat(storableFloat);
			
			if(onChange != null){
				uiSlider.slider.onValueChanged.AddListener(onChange);
			}
			
			return storableFloat;
		}
	}
}